@@ -1,4 +1,4 @@
-// Copyright (c) 2014-2018, The Monero Project
+// Copyright (c) 2014-2019, The Monero Project
 // 
 // All rights reserved.
 // 
@@ -29,6 +29,9 @@
 // Parts of this file are originally copyright (c) 2012-2013 The Cryptonote developers
 
 #pragma once
+
+#include "cryptonote_protocol/enums.h"
+
 namespace cryptonote
 {
   /************************************************************************/
@@ -36,7 +39,9 @@ namespace cryptonote
   /************************************************************************/
   struct tx_verification_context
   {
-    bool m_should_be_relayed;
+    static_assert(unsigned(relay_method::none) == 0, "default m_relay initialization is not to relay_method::none");
+
+    relay_method m_relay; // gives indication on how tx should be relayed (if at all)
     bool m_verifivation_failed; //bad tx, should drop connection
     bool m_verifivation_impossible; //the transaction is related with an alternative blockchain
     bool m_added_to_pool; 
@@ -47,7 +52,7 @@ namespace cryptonote
     bool m_too_big;
     bool m_overspend;
     bool m_fee_too_low;
-    bool m_not_rct;
+    bool m_too_few_outputs;
   };
 
   struct block_verification_context
@@ -57,5 +62,6 @@ namespace cryptonote
     bool m_marked_as_orphaned;
     bool m_already_exists;
     bool m_partial_block_reward;
+    bool m_bad_pow; // if bad pow, bad peer outright for DoS protection
   };
 }
